/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Animation
Version:   1.0
ID:        ID_ANIMATION
Category:  GUI
Date:      April 2002
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2002.  All rights reserved.
Keywords:  animation, frame, anim
Short:     Creates animations within graphical areas.

###DESCRIPTION###
<p>The Animation is a graphical class that demonstrates the power
of splitting rendered graphics into animated frames.  Unlike conventional
animation techniques where the graphics may be pre-rendered or drawn
using a paint package and saved as animated gif files, frame based
animation is calculated in real-time.  This technique uses a minimal
amount of memory and gives you the ability to animate rendered graphics
as well as images.  For example, an individual line could be be animated
to give the illusion of rotation, or a circle could expand and shrink
with a simple set-up.</p>

<p>Here is a basic animation that switches between a rectangle and circle
shape every second:</p>

<pre>
  &lt;render center width="20%" height="20%" colour="#a0a0a0"&gt;
    &lt;animation speed="1" framecount="2"/&gt;
    &lt;box width="10" height="10" border="#000000" center frame="1"&gt;
    &lt;ellipse width="10" height="10" center frame="2" border="#000000"&gt;
  &lt;/render&gt;
</pre>

<p>The two most important fields that you need to set are the Speed and
FrameCount fields.  The Speed field determines the amount of frames per
second (FPS) to be used when drawing the animation.  The FrameCount
is a compulsory setting that determines the total amount of frames included
in the animation.  In most cases, this is all that you need to worry about.</p>

<p>Any objects that you use in the animation must be able to support frames.
Most graphical objects support a Frame field as standard, but check the
documentation to make sure.  Objects that do not support frames or have not
been given a frame setting will inevitably be drawn to the graphics
container in all conditions - i.e. they will ignore the Animation object.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/functions/strtoint.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_ANIMATION 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_ANIMATION, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2002. All rights reserved.",
   "April 2002",
   "Animation"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR AnimationClass = 0;

/****************************************************************************
** Animation class.
*/

struct Animation {
   OBJECT_HEADER
   LONG     FrameCount;  /* Maximum frame count */
   FLOAT    Speed;       /* Speed of the animation */
   LONG     Frame;       /* The current frame */
   OBJECTID DrawableID;  /* Drawable that the animation is being applied to */
   LONG     NoDrawing;   /* If TRUE, the animation object will not automatically draw every frame */

   PRIVATE_FIELDS
   OBJECTID TimerID;
   FLOAT    Counter;
   MEMORYID SequenceMID;
   WORD     *Sequence;
   WORD     Position;    /* Sequence position */
};

static ERROR SET_Frame(struct Animation *, LONG *);
static ERROR SET_Speed(struct Animation *, FLOAT *);
static ERROR SET_Sequence(struct Animation *, STRING);

static struct FieldArray AnimationFields[] = {
   { "FrameCount", 0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   { "Speed",      0, FDF_FLOAT|FDF_RW,    0, NULL, SET_Speed },
   { "Frame",      0, FDF_LONG|FDF_RW,     0, NULL, SET_Frame },
   { "Drawable",   0, FDF_OBJECTID|FDF_RW, ID_RENDER, NULL, NULL },
   { "NoDrawing",  0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   /*** Virtual fields ***/
   { "Sequence",   0, FDF_STRING|FDF_W,    0, NULL, SET_Sequence },
   END_FIELD
};

static ERROR ANIM_AccessObject(struct Animation *, APTR);
static ERROR ANIM_ClosingTag(struct Animation *, APTR);
static ERROR ANIM_Free(struct Animation *, APTR);
static ERROR ANIM_NewObject(struct Animation *, APTR);
static ERROR ANIM_Init(struct Animation *, APTR);
static ERROR ANIM_ReleaseObject(struct Animation *, APTR);
static ERROR ANIM_Timer(struct Animation *, struct acTimer *);

static struct ActionArray AnimationActions[] = {
   { AC_AccessObject,  ANIM_AccessObject },
   { AC_ClosingTag,    ANIM_ClosingTag },
   { AC_Free,          ANIM_Free },
   { AC_Init,          ANIM_Init },
   { AC_NewObject,     ANIM_NewObject },
   { AC_ReleaseObject, ANIM_ReleaseObject },
   { AC_Timer,         ANIM_Timer },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&AnimationClass, NULL,
      FID_BaseClassID|TLONG, ID_ANIMATION,
      FID_Version|TFLOAT,    VER_ANIMATION,
      FID_Name|TSTRING,      "Animation",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      AnimationActions,
      FID_Fields|TPTR,       AnimationFields,
      FID_Size|TLONG,        sizeof(struct Animation),
      NULL));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (AnimationClass) { Action(AC_Free, AnimationClass, NULL); AnimationClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Animation: AccessObject
*/

static ERROR ANIM_AccessObject(struct Animation *Self, APTR Void)
{
   if (Self->SequenceMID) {
      if (AccessMemory(Self->SequenceMID, MEM_READWRITE, (void **)&Self->Sequence) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessMemory));
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Animation: ClosingTag
*/

static ERROR ANIM_ClosingTag(struct Animation *Self, APTR Void)
{
   /*** Find the Drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Animation: Free
*/

static ERROR ANIM_Free(struct Animation *Self, APTR Void)
{
   UnsubscribeTimer(Self->Head.UniqueID);

   if (Self->TimerID) {
      OBJECTPTR timer;
      if (AccessObject(Self->TimerID, 2000, &timer) IS ERR_Okay) {
         UnsubscribeAction(timer, NULL, GetUniqueID(Self));
         ReleaseObject(timer);
      }
      Self->TimerID = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Animation: Init
*/

static ERROR ANIM_Init(struct Animation *Self, APTR Void)
{
   if (Self->Frame > Self->FrameCount) Self->Frame = 1;

   return(SubscribeTimer(Self->Head.UniqueID, 10));
}

/*****************************************************************************
** Animation: NewObject
*/

static ERROR ANIM_NewObject(struct Animation *Self, APTR Void)
{
   Self->Speed      = 1.0;
   Self->Counter    = 0;
   Self->FrameCount = 1;
   Self->Frame      = 1;
   return(ERR_Okay);
}

/*****************************************************************************
** Animation: ReleaseObject
*/

static ERROR ANIM_ReleaseObject(struct Animation *Self, APTR Void)
{
   if (Self->Sequence) { ReleaseMemory(Self->Sequence); Self->Sequence = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Animation: Timer
*/

static ERROR ANIM_Timer(struct Animation *Self, struct acTimer *Args)
{
   OBJECTPTR drawable;

   Self->Counter += ((FLOAT)Args->TimeElapsed)/1000;
   if (Self->Counter >= Self->Speed) {
      Self->Counter = 0;

      if (AccessObject(Self->DrawableID, 0, &drawable) IS ERR_Okay) {
         if (Self->Sequence) {
            Self->Position++;
            if (Self->Sequence[Self->Position] IS -1) Self->Position = 0;
            Self->Frame = Self->Sequence[Self->Position];
         }
         else {
            Self->Frame++;
            if (Self->Frame > Self->FrameCount) Self->Frame = 1;
         }

         SetField(drawable, FID_Frame, FT_LONG, &Self->Frame);
         if (Self->NoDrawing IS FALSE) Action(AC_Draw, drawable, NULL);
         ReleaseObject(drawable);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  Refers to the drawable that the Animation is being applied to.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In most cases, when you create a new Animation it should be contained by a
drawable area, such as a Render object.  If you do this correctly then
the Animation will detect the drawable area and set this field to the
drawable object. However, there may be occasions where you want to
initialise a Animation object to a non-graphical owner.  If this is the
case then you must set the Drawable field to point to a valid drawable
area, such as a Render object.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frame
Type:   LONG
Status: Read/Write
Short:  Reflects the current frame number.

###DESCRIPTION###
<p>The current frame number is always reflected in this field.  It can
sometimes be desirable to set this field on initialisation to force the
starting frame number.  The range of the Frame value is from 0 to one less
than the value in the FrameCount field.</p>

###SEE ALSO###
Field: FrameCount

###END###

*****************************************************************************/

static ERROR SET_Frame(struct Animation *Self, LONG *Value)
{
   if (*Value < 0) return(ERR_Failed);
   Self->Frame = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   FrameCount
Type:   LONG
Status: Read/Write
Short:  The total number of frames being used in the animation.

###DESCRIPTION###
<p>This field must be set to the total number of frames that you want to use
in the animation.  The minimum FrameCount that you can set is 1.</p>

###SEE ALSO###
Field: Frame

###END###

*****************************************************************************/

static ERROR SET_FrameCount(struct Animation *Self, LONG *Value)
{
   if ((*Value < 1) OR (*Value > 10000)) return(ERR_Failed);

   Self->FrameCount = *Value;

   /*** Check that the Frame number is still less than the FrameCount ***/

   if (Self->Frame >= Self->FrameCount) Self->Frame = 1;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Sequence
Type:   STRING
Status: Set
Short:  Detailed frame sequencing can be set in this field.

###DESCRIPTION###
<p>Normally an animation object will follow a simple frame sequence, starting
from frame 1 and ending at the frame determined by the FrameCount field.  You
can change this default behaviour and create a </p>

###END###

*****************************************************************************/

static ERROR SET_Sequence(struct Animation *Self, STRING Value)
{
   STRING str;
   LONG total, pos, frame, end;

   if ((!Value) OR (!(*Value))) return(ERR_Args);

   /*** Calculate the size of the sequencing array that we are going to need ***/

   total = 0;
   for (str=Value; *str; str++) {
      if ((*str >= '0') AND (*str <= '9')) {
         frame = StrToInt(str);
         while ((*str >= '0') AND (*str <= '9')) str++;
         if (frame > total) total = frame;
      }
      else str++;
   }

   /*** Allocate the sequencing array and record the sequence ***/

   if (AllocMemory((total+1) * sizeof(WORD), MEM_NOCLEAR|Self->Head.MemFlags, &Self->Sequence, &Self->SequenceMID) IS ERR_Okay) {
      pos = 0;
      for (str=Value; *str; str++) {
         if ((*str >= '0') AND (*str <= '9')) {
            frame = StrToInt(str);
            while ((*str >= '0') AND (*str <= '9')) str++;
            while ((*str) AND (*str <= 0x20)) str++;
            if (*str IS '-') {
               while ((*str) AND (*str < '0') AND (*str > '9')) str++;
               end = StrToInt(str);
               while (frame != end) {
                  Self->Sequence[pos++] = frame;
                  if (frame < end) frame++;
                  else frame--;
               }
            }
            else Self->Sequence[pos++] = frame;
         }
         else str++;
      }

      Self->Sequence[pos] = -1; /* Sequence termination */
   }
   else return(ERR_AllocMemory);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Speed
Type:   FLOAT
Status: Read/Write
Short:  Sets the speed of the animation in frames per second.

###DESCRIPTION###
<p>The speed of the animation may be set with this field.  The setting must
be calculated in frames per second - so a value of 1 would change frames as
every second passes.  The maximum value is 100 frames per second, but it is
recommended that 25 frames per second is specified for most purposes.</p>

###END###

*****************************************************************************/

static ERROR SET_Speed(struct Animation *Self, FLOAT *Value)
{
   if (*Value > 100) Self->Speed = 100;
   else if (*Value < 0) Self->Speed = 0;
   else Self->Speed = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

                 "I lurk in the alleys, wait for the kill
                  I have no remorse for the blood that I spill.
                  A merciless butcher, who lives underground
                  I'm out to destroy, and I will cut you down."

                       - Megadeth

*****************************************************************************/
