/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Image
Version:   1.0
ID:        ID_IMAGE
Category:  GUI
Date:      June 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Keywords:  image, picture, draw, graphic, portrait, art, artwork
Short:     Draws images to drawable areas.

###DESCRIPTION###
<p>The Image class is designed for object scripting and closely follows
the general definition of the 'img' tag in HTML.  The main difference that
you will notice is that it sports a few extra effects, and you have complete
control over the positioning of the image graphic.</p>

<p>Here is a DML based example for loading a picture file via an Image
object:</p>

<pre>
  &lt;image src="pictures:logos/athene.jpeg" x="25" y="50"/&gt;
</pre>

<p>The Image class supports transparent colour management for file formats
such as GIF and PNG, as well as images that are accompanied by an alpha
channel.  If an image utilises a transparent colour but the file format
cannot declare it, you can tell the object what the mask colour is.  This
is set via the <a href="#Mask">Mask</a> field, e.g:</p>

<pre>
  &lt;image src="icons:misc/paintcan.pcx" mask="128,0,0" width="32" height="32"/&gt;
</pre>

<p>There are no restrictions on the data format of the picture file, but it
must be supported by one of the Picture classes in the system.  If for example
the file format is JPEG, but the system does not have a JPEG Picture class
installed, it will not be possible to load the file.  To obtain a list of
supported file formats, check the list of Picture classes in the Graphics
category.</p>

<p>All pictures that are loaded via the Image class are cached into a shared
memory pool.  This means that if a particular image file is loaded multiple
times, the data will be stored only once to save on memory and load times.
Image files are automatically unloaded when their open-count reaches zero.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/functions/strtoint.c>
#include <pandora/modules/render.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_IMAGE 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_IMAGE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003.  All rights reserved.",
   "October 2003",
   "Image"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR ImageClass = 0;
static OBJECTPTR modRender = NULL;
static struct RenderBase *RenderBase;
static struct ModuleMaster *ImageModule = 0;

/****************************************************************************
** Image class.
*/

struct Image {
   OBJECT_HEADER
   LONG     AlignFlags;
   LONG     Frame;
   LONG     Flags;
   OBJECTID DrawableID;
   LONG     Visible;
   LONG     ImageWidth;
   LONG     ImageHeight;
   LONG     ImageX;
   LONG     ImageY;

   PRIVATE_FIELDS
   struct Picture *Picture;
   MEMORYID LocationMID;
   LONG     Dimensions;
   WORD     Opacity;
   FLOAT    XCoord;
   FLOAT    YCoord;
   FLOAT    YOffset;
   FLOAT    XOffset;
   FLOAT    Width;
   FLOAT    Height;
   STRING   Location;
   struct RGB MaskRGB;
   struct RGB BackgroundRGB;
   WORD TotalFrames;
   WORD *Frames;
   MEMORYID FramesMID;
};

static struct FieldDef AlignFlags[] = {
   { "RIGHT",    ALIGN_RIGHT    }, { "LEFT",       ALIGN_LEFT    },
   { "BOTTOM",   ALIGN_BOTTOM   }, { "TOP",        ALIGN_TOP     },
   { "CENTER",   ALIGN_CENTER   }, { "MIDDLE",     ALIGN_MIDDLE  },
   { "VERTICAL", ALIGN_VERTICAL }, { "HORIZONTAL", ALIGN_HORIZONTAL },
   { NULL, NULL }
};

#define IMF_MASK       0x00000001
#define IMF_TILE       0x00000002
#define IMF_ENLARGE    0x00000008
#define IMF_SHRINK     0x00000010
#define IMF_CENTER     0x00000020
#define IMF_SOLIDBLEND 0x00000040
#define IMF_BACKGROUND 0x00000080
#define IMF_STICKY     0x00000100
#define IMF_NOFAIL     0x00000200
#define IMF_11RATIO    0x00000400
#define IMF_FIT        0x00000800

#define IMF_STRETCH  (IMF_ENLARGE|IMF_SHRINK)

static struct FieldDef ImageFlags[] = {
   { "MASK",      IMF_MASK },    { "TILE",       IMF_TILE },
   { "ENLARGE",    IMF_ENLARGE },
   { "SHRINK",    IMF_SHRINK },  { "STRETCH",    IMF_STRETCH },
   { "CENTER",    IMF_CENTER },  { "SOLIDBLEND", IMF_SOLIDBLEND },
   { "STICKY",    IMF_STICKY },  { "NOFAIL",     IMF_NOFAIL },
   { "11RATIO",   IMF_11RATIO }, { "FIT",        IMF_FIT },
   { NULL, NULL }
};

static ERROR GET_Height(struct Image *, struct Variable *);
static ERROR GET_Location(struct Image *, STRING *);
static ERROR GET_MaskRGB(struct Image *, struct RGB *);
static ERROR GET_Opacity(struct Image *, FLOAT *);
static ERROR GET_PixelSize(struct Image *, LONG *);
static ERROR GET_Tile(struct Image *, LONG *);
static ERROR GET_Width(struct Image *, struct Variable *);
static ERROR GET_X(struct Image *, struct Variable *);
static ERROR GET_XOffset(struct Image *, struct Variable *);
static ERROR GET_Y(struct Image *, struct Variable *);
static ERROR GET_YOffset(struct Image *, struct Variable *);

static ERROR SET_Background(struct Image *, STRING);
static ERROR SET_Flags(struct Image *, LONG *);
static ERROR SET_Frames(struct Image *, STRING);
static ERROR SET_Location(struct Image *, STRING);
static ERROR SET_Height(struct Image *, struct Variable *);
static ERROR SET_Mask(struct Image *, STRING);
static ERROR SET_MaskRGB(struct Image *, struct RGB *);
static ERROR SET_Opacity(struct Image *, FLOAT *);
static ERROR SET_Tile(struct Image *, LONG *);
static ERROR SET_Width(struct Image *, struct Variable *);
static ERROR SET_X(struct Image *, struct Variable *);
static ERROR SET_XOffset(struct Image *, struct Variable *);
static ERROR SET_Y(struct Image *, struct Variable *);
static ERROR SET_YOffset(struct Image *, struct Variable *);

static struct FieldArray ImageFields[] = {
   { "Align",       0, FDF_LONGFLAGS|FDF_RW, (LONG)&AlignFlags, NULL, NULL },
   { "Frame",       0, FDF_LONG|FDF_RW,      0,                 NULL, NULL },
   { "Flags",       0, FDF_LONGFLAGS|FDF_RW, (LONG)&ImageFlags, NULL, SET_Flags },
   { "Drawable",    0, FDF_OBJECTID|FDF_RI,  0,                 NULL, NULL },
   { "Visible",     0, FDF_LONG|FDF_RW,      0,                 NULL, NULL },
   { "ImageWidth",  0, FDF_LONG|FDF_RI,      0,                 NULL, NULL },
   { "ImageHeight", 0, FDF_LONG|FDF_RI,      0,                 NULL, NULL },
   { "ImageX",      0, FDF_LONG|FDF_RI,      0,                 NULL, NULL },
   { "ImageY",      0, FDF_LONG|FDF_RI,      0,                 NULL, NULL },
   /*** Virtual Fields ***/
   { "Background",  0, FDF_STRING|FDF_W,   0, NULL,          SET_Background },
   { "Frames",      0, FDF_STRING|FDF_W,   0, NULL,          SET_Frames },
   { "Location",    0, FDF_STRING|FDF_RW,  0, GET_Location,  SET_Location },
   { "Mask",        0, FDF_STRING|FDF_W,   0, NULL,          SET_Mask },
   { "MaskRGB",     0, FDF_POINTER|FDF_RW, 0, GET_MaskRGB,   SET_MaskRGB },
   { "Opacity",     0, FDF_FLOAT|FDF_RW,   0, GET_Opacity,   SET_Opacity },
   { "PixelSize",   0, FDF_LONG|FDF_R,     0, GET_PixelSize, NULL },
   { "Src",         0, FDF_STRING|FDF_RW,  0, GET_Location,  SET_Location },
   { "Tile",        0, FDF_LONG|FDF_RI,    0, GET_Tile,      SET_Tile },
   /*** Variable Fields ***/
   { "Height",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

static ERROR IMAGE_AccessObject(struct Image *, APTR);
static ERROR IMAGE_Free(struct Image *, APTR);
static ERROR IMAGE_NewObject(struct Image *, APTR);
static ERROR IMAGE_Hide(struct Image *, APTR);
static ERROR IMAGE_Init(struct Image *, APTR);
static ERROR IMAGE_Move(struct Image *, struct acMove *);
static ERROR IMAGE_MoveToPoint(struct Image *, struct acMoveToPoint *);
static ERROR IMAGE_ReleaseObject(struct Image *, APTR);
static ERROR IMAGE_ScrollToPoint(struct Image *, struct acScrollToPoint *);
static ERROR IMAGE_Show(struct Image *, APTR);

static struct ActionArray ImageActions[] = {
   { AC_AccessObject,  IMAGE_AccessObject },
   { AC_Free,          IMAGE_Free },
   { AC_NewObject,     IMAGE_NewObject },
   { AC_Hide,          IMAGE_Hide },
   { AC_Init,          IMAGE_Init },
   { AC_Move,          IMAGE_Move },
   { AC_MoveToPoint,   IMAGE_MoveToPoint },
   { AC_ReleaseObject, IMAGE_ReleaseObject },
   { AC_ScrollToPoint, IMAGE_ScrollToPoint },
   { AC_Show,          IMAGE_Show },
   { NULL, NULL }
};

static LONG GetX(struct Image *, struct Render *);
static LONG GetY(struct Image *, struct Render *);
static LONG GetHeight(struct Image *, struct Render *);
static LONG GetWidth(struct Image *, struct Render *);

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   GetField(argModule, FID_Master, FT_POINTER, &ImageModule);

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modRender, NULL, FID_Name|TSTRING, "render", TAGEND) IS ERR_Okay) {
      if (GetField(modRender, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ImageClass, NULL,
      FID_BaseClassID|TLONG, ID_IMAGE,
      FID_Name|TSTRING,      "Image",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      ImageActions,
      FID_Fields|TPTR,       ImageFields,
      FID_Size|TLONG,        sizeof(struct Image),
      FID_Flags|TLONG,       CLF_PRIVATEONLY,
      FID_Version|TFLOAT,    VER_IMAGE,
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (ImageClass) { acFree(ImageClass); ImageClass = 0; }
   if (modRender)  { acFree(modRender);  modRender = 0; }
   return(ERR_Okay);
}

/*****************************************************************************
** Image: AccessObject
*/

static ERROR IMAGE_AccessObject(struct Image *Self, APTR Void)
{
   if (Self->FramesMID) {
      if (AccessMemory(Self->FramesMID, MEM_READWRITE, &Self->Frames) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessMemory));
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: DrawImage()
*/

static void DrawImage(struct Image *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct mtBlitArea area;
   struct Bitmap *picbitmap;
   struct Picture *picture;
   struct ClipRectangle clip;
   WORD opacity;
   FLOAT ratio, vratio;
   LONG width, height, x, y, xdest, ydest;
   LONG imagex, imagey, imagewidth, imageheight;
      
   if (Self->Visible IS FALSE) return;

   /* Determine whether or not we need to draw this image
   ** based on our object's frame settings.
   */

   if (Self->Frame) {
      if (Drawable->Frame != Self->Frame) return;
   }
   else if (Self->Frames) {
      if ((Drawable->Frame > Self->TotalFrames) OR (Self->Frames[Drawable->Frame] < 1)) return;
   }

   if ((picture = Self->Picture)) {
      GetField(picture, FID_Bitmap, FT_POINTER, &picbitmap);
   }

   xdest = GetX(Self, Drawable);
   ydest = GetY(Self, Drawable);
   width = GetWidth(Self, Drawable);
   height = GetHeight(Self, Drawable);

   if (!picture) {
      /* If no picture is available, the only thing that we are capable of
      ** doing is clearing the background.
      */

      if (Self->Flags & IMF_BACKGROUND) {
         ActionTags(MT_DrawRGBRectangle, Bitmap, xdest, ydest, width, height,
           Self->BackgroundRGB.Red, Self->BackgroundRGB.Green, Self->BackgroundRGB.Blue, TRUE);
      }

      return;
   }

   /* Activate alpha-blending to the destination if the picture supports
   ** an alpha channel and our blend flag has been set.
   */

   if ((!(Self->Flags & IMF_SOLIDBLEND)) AND (picbitmap->Flags & BMF_ALPHA)) {
      picbitmap->Flags |= BMF_ALPHABLEND;
   }

   clip.left   = Bitmap->ClipLeft;
   clip.top    = Bitmap->ClipTop;
   clip.right  = Bitmap->ClipRight;
   clip.bottom = Bitmap->ClipBottom;
   if (Bitmap->ClipLeft   < xdest) Bitmap->ClipLeft = xdest;
   if (Bitmap->ClipTop    < ydest) Bitmap->ClipTop  = ydest;
   if (Bitmap->ClipRight  > xdest + width)  Bitmap->ClipRight  = xdest + width;
   if (Bitmap->ClipBottom > ydest + height) Bitmap->ClipBottom = ydest + height;

   opacity = picbitmap->Opacity;
   picbitmap->Opacity = Self->Opacity;

   imagewidth = Self->ImageWidth;
   imageheight = Self->ImageHeight;

   /*** If stretching is enabled then we will adjust the image to match the size of the drawable space ***/

   if (Self->Flags & IMF_ENLARGE) {
      if (Self->Flags & IMF_11RATIO) {
         ratio = (FLOAT)width / (FLOAT)picbitmap->Width;
         vratio = (FLOAT)height / (FLOAT)picbitmap->Height;

         if (Self->Flags & IMF_FIT) {
            /*** Enlarge the image to fit the display ***/
            if (ratio > vratio) ratio = vratio;
         }
         else {
            /*** Enlarge the image to cover the entirety of the display ***/
            if (ratio < vratio) ratio = vratio;
         }

         if (ratio >= 1.0) {
            imagewidth  = picbitmap->Width * ratio;
            imageheight = picbitmap->Height * ratio;
         }
      }
      else {
         if (imagewidth < width) imagewidth = width;
         if (imageheight < height) imageheight = height;
      }
   }

   if (Self->Flags & IMF_SHRINK) {
      if (Self->Flags & IMF_11RATIO) {
         ratio = (FLOAT)width / (FLOAT)picbitmap->Width;
         vratio = (FLOAT)height / (FLOAT)picbitmap->Height;

         if (Self->Flags & IMF_FIT) {
            /*** Shrink the image to fit the display ***/
            if (ratio > vratio) ratio = vratio;
         }
         else {
            /*** Shrink the image to cover the entirety of the display ***/
            if (ratio < vratio) ratio = vratio;
         }

         if (ratio <= 1.0) {
            imagewidth  = picbitmap->Width * ratio;
            imageheight = picbitmap->Height * ratio;
         }
      }
      else {
         if (imagewidth > width) imagewidth = width;
         if (imageheight > height) imageheight = height;
      }
   }

   /*** Determine the position of the image ***/

   if ((Self->AlignFlags & ALIGN_HORIZONTAL) OR (Self->Flags & IMF_STRETCH)) imagex = xdest + ((width - imagewidth) / 2);
   else if (Self->AlignFlags & ALIGN_RIGHT) imagex = xdest + (width  - imagewidth);
   else imagex = Self->ImageX + xdest;

   if ((Self->AlignFlags & ALIGN_VERTICAL) OR (Self->Flags & IMF_STRETCH)) imagey = ydest + ((height - imageheight) / 2);
   else if (Self->AlignFlags & ALIGN_BOTTOM) imagey = ydest + (height  - imageheight);
   else imagey = Self->ImageY + ydest;

   /*** Draw the image ***/

   if (Self->Flags & IMF_TILE) {
      area.DestBitmap = Bitmap;
      area.Flags      = NULL;
      area.XCoord     = 0;
      area.YCoord     = 0;
      for (y=ydest; (y < ydest + height); y += Self->ImageHeight) {
         area.YDest  = y;
         area.Height  = Self->ImageHeight;
         if (area.YDest + area.Height > ydest + height) area.Height = (ydest + height) - y;

         for (x=xdest; (x < xdest + width); x += Self->ImageWidth) {
            area.XDest  = x;
            area.Width  = Self->ImageWidth;
            if (x + Self->ImageWidth > xdest + width) area.Width = (xdest + width) - x;
            Action(MT_BlitArea, picbitmap, &area);
         }
      }
   }
   else {
      if (Self->Flags & IMF_BACKGROUND) {
         if ((picbitmap->Flags & (BMF_ALPHABLEND|BMF_TRANSPARENT)) OR
             ((picbitmap->Width < 64) AND (picbitmap->Height < 64))) {
            ActionTags(MT_DrawRGBRectangle, Bitmap, xdest, ydest, width, height,
              Self->BackgroundRGB.Red, Self->BackgroundRGB.Green, Self->BackgroundRGB.Blue, TRUE);
         }
         else {
            /*** Top ***/
            if (imagey > ydest) {
               ActionTags(MT_DrawRGBRectangle, Bitmap, xdest, ydest, width, imagey - ydest,
                 Self->BackgroundRGB.Red, Self->BackgroundRGB.Green, Self->BackgroundRGB.Blue, TRUE);
            }

            /*** Bottom ***/
            if (imagey + imageheight < ydest + height) {
               ActionTags(MT_DrawRGBRectangle, Bitmap, xdest, imagey + imageheight,
                 width, (ydest + height) - (imagey + imageheight),
                 Self->BackgroundRGB.Red, Self->BackgroundRGB.Green, Self->BackgroundRGB.Blue, TRUE);
            }

            /*** Left ***/
            if (imagex > xdest) {
               ActionTags(MT_DrawRGBRectangle, Bitmap, xdest, imagey, imagex - xdest, imageheight,
                 Self->BackgroundRGB.Red, Self->BackgroundRGB.Green, Self->BackgroundRGB.Blue, TRUE);
            }

            /*** Right ***/
            if (imagex + imagewidth < xdest + width) {
               ActionTags(MT_DrawRGBRectangle, Bitmap, imagex + imagewidth, imagey,
                 (xdest + width) - (imagex + imagewidth), imageheight,
                 Self->BackgroundRGB.Red, Self->BackgroundRGB.Green, Self->BackgroundRGB.Blue, TRUE);
            }
         }
      }

      if ((imagewidth != picbitmap->Width) OR (imageheight != picbitmap->Height)) {
         ActionTags(MT_BlitStretch, picbitmap, Bitmap, BSF_BILINEAR, 0, 0,
            picbitmap->Width, picbitmap->Height, imagex, imagey, imagewidth, imageheight);
      }
      else {
         ActionTags(MT_BlitArea, picbitmap, Bitmap, NULL, 0, 0,
            imagewidth, imageheight, imagex, imagey);
      }
   }

   picbitmap->Opacity = opacity;
   picbitmap->Flags &= ~BMF_ALPHABLEND;

   Bitmap->ClipLeft   = clip.left;
   Bitmap->ClipTop    = clip.top;
   Bitmap->ClipRight  = clip.right;
   Bitmap->ClipBottom = clip.bottom;
}

/*****************************************************************************
** Image: Free
*/

static ERROR IMAGE_Free(struct Image *Self, APTR Void)
{
   OBJECTPTR drawable;

   if (Self->Location)    { ReleaseMemory(Self->Location);   Self->Location    = NULL; }
   if (Self->LocationMID) { FreeMemoryID(Self->LocationMID); Self->LocationMID = NULL; }

   if (Self->Frames)    { ReleaseMemory(Self->Frames);   Self->Frames    = NULL; }
   if (Self->FramesMID) { FreeMemoryID(Self->FramesMID); Self->FramesMID = NULL; }

   if (Self->Picture) {
      Action(AC_Free, Self->Picture, NULL);
      Self->Picture = NULL;
   }

   if (Self->DrawableID) {
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, drawable, Self, NULL);
         ReleaseObject(drawable);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Hides an image from view.
###END###

*****************************************************************************/

static ERROR IMAGE_Hide(struct Image *Self, APTR Void)
{
   struct acDraw draw;

   if (Self->Visible IS TRUE) {
      Self->Visible = FALSE;
      GetFields(Self, FID_XCoord|TDOUBLE, &draw.XCoord,
                      FID_YCoord|TDOUBLE, &draw.YCoord,
                      FID_Width|TDOUBLE,  &draw.Width,
                      FID_Height|TDOUBLE, &draw.Height,
                      TAGEND);
      ActionMsg(AC_Draw, Self->DrawableID, &draw);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Image: Init
*/

static ERROR IMAGE_Init(struct Image *Self, APTR Void)
{
   struct RenderInfo info;
   struct Picture *picture;
   struct Variable value;
   OBJECTPTR drawable;
   OBJECTID OwnerID;
   LONG ContainerWidth, ContainerHeight;
   ERROR error;

   if ((!Self->LocationMID) AND (!(Self->Flags & IMF_NOFAIL))) return(ObjectError(Self, ERH_Init, ERR_MissingLocation));

   if (!Self->Location) {
      if (GET_Location(Self, &Self->Location) != ERR_Okay) {
         if (!(Self->Flags & IMF_NOFAIL)) return(ObjectError(Self, ERH_Init, ERR_GetField));
      }
   }
   
   /*** Find the Drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   error = ERR_Okay;
   if (!Self->Picture) {
      if (NewObject(ID_PICTURE, NULL, &picture, NULL) IS ERR_Okay) {

         if (GetRenderInfo(Self->DrawableID, &info, sizeof(info)) IS ERR_Okay) {
            picture->Bitmap->BitsPerPixel = info.BitsPerPixel;
            picture->Flags |= PCF_32BITALPHA|PCF_CACHE;
            SetField(picture, FID_Location, FT_STRING, Self->Location);

            /*** If IMF_MASK is set, force our preset mask colour on the picture ***/

            if (Self->Flags & IMF_MASK) {
               picture->Flags |= PCF_BACKGROUND;
               picture->Bitmap->ColourRGB->Red   = Self->MaskRGB.Red;
               picture->Bitmap->ColourRGB->Green = Self->MaskRGB.Green;
               picture->Bitmap->ColourRGB->Blue  = Self->MaskRGB.Blue;
            }

            if (Action(AC_Init, picture, NULL) IS ERR_Okay) {
               Self->Picture = picture;

               /*** Retrieve the mask colour from the loaded picture ***/

               Self->MaskRGB.Red   = picture->Bitmap->ColourRGB->Red;
               Self->MaskRGB.Green = picture->Bitmap->ColourRGB->Green;
               Self->MaskRGB.Blue  = picture->Bitmap->ColourRGB->Blue;
            }
            else {
               DPrintF("@Image:","Failed to read picture \"%s\".", Self->Location);
               error = ERR_Init;
            }

            if (error != ERR_Okay) acFree(picture);
         }
         else if (!(Self->Flags & IMF_NOFAIL)) return(ObjectError(Self, ERH_Init, ERR_AccessObject));
      }
      else if (!(Self->Flags & IMF_NOFAIL)) return(ObjectError(Self, ERH_Init, ERR_NewObject));
   }

   if ((error != ERR_Okay) AND (!(Self->Flags & IMF_NOFAIL))) return(error);

   if ((!Self->Picture) AND (!(Self->Flags & IMF_NOFAIL))) {
      return(ERR_Failed);
   }

   /* If dimension settings are missing (e.g. if it is impossible to
   ** determine width, height or a coordinate), then we set the missing
   ** fields to maximum possible values.
   */

   value.Type = FD_DOUBLE;
   value.Double = 0;

   if (!(Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_X(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_Y(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_XOffset(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_YOffset(Self, &value);
   }

   /*** Continue processing ***/

   if (AccessObject(Self->DrawableID, 5000, &drawable) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_AccessObject));
   }

   error = ERR_Okay;
   if (Self->Picture) {
      picture = Self->Picture;
      if (!Self->ImageWidth)  Self->ImageWidth  = picture->Bitmap->Width;
      if (!Self->ImageHeight) Self->ImageHeight = picture->Bitmap->Height;

      if (Self->Flags & IMF_ENLARGE) {
         /*** Enlarge the image to fit the owner ***/
         if (GetField(Self, FID_Width, FT_LONG, &ContainerWidth) IS ERR_Okay) {
            if (Self->ImageWidth < ContainerWidth) {
               DPrintF("Image:","The width will be enlarged to fit.");
               Self->ImageWidth = ContainerWidth;
            }
         }
         if (GetField(Self, FID_Height, FT_LONG, &ContainerHeight) IS ERR_Okay) {
            if (Self->ImageHeight < ContainerHeight) {
               DPrintF("Image:","The height will be enlarged to fit.");
               Self->ImageHeight = ContainerHeight;
            }
         }
      }

      if (Self->Flags & IMF_SHRINK) {
         /*** Shrink the image to fit the owner ***/
         if (GetField(Self, FID_Width, FT_LONG, &ContainerWidth) IS ERR_Okay) {
            if (Self->ImageWidth > ContainerWidth) {
               DPrintF("Image:","The width will be shrunk to fit.");
               Self->ImageWidth = ContainerWidth;
            }
         }
         if (GetField(Self, FID_Height, FT_LONG, &ContainerHeight) IS ERR_Okay) {
            if (Self->ImageHeight > ContainerHeight) {
               DPrintF("Image:","The height will be shrunk to fit.");
               Self->ImageHeight = ContainerHeight;
            }
         }
      }
   }

   ActionTags(MT_SetRenderCallback, drawable, Self, &DrawImage);

   ReleaseObject(drawable);
   return(error);
}

/*****************************************************************************
** Image: Move
*/

static ERROR IMAGE_Move(struct Image *Self, struct acMove *Args)
{
   if (!Args) return(ERR_Args);

   if (Self->Flags & IMF_STICKY) return(ERR_Okay);

   if ((!Args->XChange) AND (!Args->YChange)) return(ERR_Okay);

   Self->XCoord -= Args->XChange;
   Self->YCoord -= Args->YChange;

   Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;

   ActionMsg(AC_Draw, Self->DrawableID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************
** Image: MoveToPoint
*/

static ERROR IMAGE_MoveToPoint(struct Image *Self, struct acMoveToPoint *Args)
{
   LONG oldx, oldy;

   if (!Args) return(ERR_Args);

   if (Self->Flags & IMF_STICKY) return(ERR_Okay);

   oldx = Self->XCoord;
   oldy = Self->YCoord;

   if (Args->Flags & MTF_X) {
      Self->XCoord = Args->XCoord;
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   }

   if (Args->Flags & MTF_Y) {
      Self->YCoord = Args->YCoord;
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
   }

   if ((oldx != Self->XCoord) OR (oldy != Self->YCoord)) ActionMsg(AC_Draw, Self->DrawableID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************
** Image: NewObject
*/

static ERROR IMAGE_NewObject(struct Image *Self, APTR Void)
{
   Self->Visible = TRUE;
   Self->Opacity = 255;
   return(ERR_Okay);
}

/*****************************************************************************
** Image: ReleaseObject
*/

static ERROR IMAGE_ReleaseObject(struct Image *Self, APTR Void)
{
   if (Self->Frames)   { ReleaseMemory(Self->Frames); Self->Frames = NULL; }
   if (Self->Location) { ReleaseMemory(Self->Location); Self->Location = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  ScrollToPoint
Short: Scrolls an image within its allocated drawing space.
###END###

*****************************************************************************/

static ERROR IMAGE_ScrollToPoint(struct Image *Self, struct acScrollToPoint *Args)
{
   struct mtMoveContent move;
   struct Render *drawable;
   LONG oldx, oldy, xchange, ychange, x, y;
   LONG areax, areay, areawidth, areaheight;

   if (!Args) return(ObjectError(Self, ERH_ScrollToPoint, ERR_Args));

   if ((Args->XCoord IS Self->ImageX) AND (Args->YCoord IS Self->ImageY)) return(ERR_Okay);

   if (AccessObject(Self->DrawableID, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
      areax = GetX(Self, drawable);
      areay = GetY(Self, drawable);
      areawidth  = GetWidth(Self, drawable);
      areaheight = GetHeight(Self, drawable);

      if (Args->Flags & STF_X) x = -Args->XCoord;
      else x = Self->ImageX;

      if (Args->Flags & STF_Y) y = -Args->YCoord;
      else y = Self->ImageY;

      oldx = Self->ImageX;
      oldy = Self->ImageY;

      Self->ImageX = x;
      Self->ImageY = y;

      xchange = Self->ImageX - oldx;
      ychange = Self->ImageY - oldy;

      if ((xchange >= areawidth-10) OR (ychange >= areaheight-10) OR
          (xchange <= -areawidth+10) OR (ychange <= -areaheight+10)) {
         Action(AC_Draw, drawable, NULL);
      }
      else {
         /*** Set horizontal movement parameters ***/

         if (x < oldx) {
            move.XCoord = areax - xchange;
            move.Width  = areawidth;
            move.XDest  = areax;
         }
         else {
            move.XCoord = areax - xchange;
            move.Width  = areawidth;
            move.XDest  = areax;
         }

         /*** Set vertical movement parameters ***/

         if (y < oldy) {
            move.YCoord = areay - ychange;
            move.Height = areaheight;
            move.YDest  = areay;
         }
         else {
            move.YCoord = areay - ychange;
            move.Height = areaheight;
            move.YDest  = areay;
         }

         move.ClipLeft   = areax;
         move.ClipTop    = areay;
         move.ClipRight  = areax + areawidth;
         move.ClipBottom = areay + areaheight;

         move.Flags  = MCF_EXPOSE;
         Action(MT_MoveContent, drawable, &move);
      }

      ReleaseObject(drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Shows an image.
###END###

*****************************************************************************/

static ERROR IMAGE_Show(struct Image *Self, APTR Void)
{
   struct acDraw draw;

   if (Self->Visible IS FALSE) {
      Self->Visible = TRUE;
      GetFields(Self, FID_XCoord|TDOUBLE, &draw.XCoord,
                      FID_YCoord|TDOUBLE, &draw.YCoord,
                      FID_Width|TDOUBLE,  &draw.Width,
                      FID_Height|TDOUBLE, &draw.Height,
                      TAGEND);
      ActionMsg(AC_Draw, Self->DrawableID, &draw);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Function: ReadByte()
** Short:    Reads any double digit hex value and converts it into a number.
*/

static WORD ReadByte(STRING str)
{
   WORD Value;
   if (!*str) return(0);
   if ((*str >= '0') AND (*str <= '9')) Value = (*str - '0') * 16;
   else if ((*str >= 'A') AND (*str <= 'F')) Value = ((*str - 'A')+10) * 16;
   else if ((*str >= 'a') AND (*str <= 'f')) Value = ((*str - 'a')+10) * 16;
   else Value = 0;
   str++;
   if ((*str >= '0') AND (*str <= '9')) Value += (*str - '0');
   else if ((*str >= 'A') AND (*str <= 'F')) Value += (*str - 'A') + 10;
   else if ((*str >= 'a') AND (*str <= 'f')) Value += (*str - 'a') + 10;
   return(Value);
}

/*****************************************************************************

###FIELD###
Name:   Align
Short:  This field allows you to align an image within its owner.
Type:   LONG/FLAGS
Prefix: ALIGN
Status: Read/Write

###DESCRIPTION###
<p>If you would like to set an abstract position for an image, you can
give it an alignment.  Note that setting the alignment over-rides any
settings in related coordinate fields.  Valid alignment flags are as
follows:</p>

<blockquote>
<p>BOTTOM<br>
CENTER/MIDDLE<br>
LEFT<br>
HORIZONTAL<br>
RIGHT<br>
TOP<br>
VERTICAL</p>
</blockquote>

###END###

*****************************************************************************/

/*****************************************************************************

*****************************************************************************/

static ERROR SET_Background(struct Image *Self, STRING Background)
{
   if (!Background) return(ObjectError(Self, ERH_SetField, ERR_NoData));

   if (*Background IS '#') {
      Background++;
      Self->BackgroundRGB.Red = ReadByte(Background);
      if (*Background) Background++; if (*Background) Background++;
      Self->BackgroundRGB.Green = ReadByte(Background);
      if (*Background) Background++; if (*Background) Background++;
      Self->BackgroundRGB.Blue = ReadByte(Background);
   }
   else {
      Self->BackgroundRGB.Red = StrToInt(Background);
      while ((*Background) AND (*Background != ',')) Background++; if (*Background) Background++;
      Self->BackgroundRGB.Green = StrToInt(Background);
      while ((*Background) AND (*Background != ',')) Background++; if (*Background) Background++;
      Self->BackgroundRGB.Blue = StrToInt(Background);
   }
   Self->Flags |= IMF_BACKGROUND;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Special flags.
Type:   LONG/FLAGS
Prefix: IMF
Status: Read/Write

###DESCRIPTION###
<p>Special flags that are available for the Image class are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">CENTER</td><td>This is a helper flag that provides an alternative to centering an image with the Align field.</td></tr>
<tr><td width="10%">ENLARGE&nbsp;&nbsp;</td><td>If the dimensions of the original picture are smaller than the destination drawable, the image will be stretched to fit.  Enlargement is dynamic, so if the drawable changes in size then the size of the picture will be recalculated to fit.</td></tr>
<tr><td width="10%">MASK</td><td>This flag will be set automatically when the Mask or MaskRGB fields are written.  You can turn off masking by switching the flag off.</td></tr>
<tr><td width="10%">SHRINK</td><td>If the dimensions of the original picture are larger than the destination drawable, the image will be stretched to fit.  Shrinking is dynamic, so if the drawable changes in size then the size of the picture will be recalculated to fit.</td></tr>
<tr><td width="10%">TILE</td><td>If an image is to be tiled, you can set this flag as an alternative to setting the Tile field.  You can turn off tiling by switching the flag off.</td></tr>
<tr><td width="10%">STRETCH</td><td>This flag conveniently combines the ENLARGE and SHRINK flags to ensure that the image always covers the dimensions of the destination drawable.</td></tr>
</table>

###END###

*****************************************************************************/

static ERROR SET_Flags(struct Image *Self, LONG *Value)
{
   Self->Flags = (Self->Flags & 0xffff0000) | (*Value & 0x0000ffff);
   if (Self->Flags & IMF_CENTER) Self->AlignFlags = ALIGN_CENTER;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Determines the frame that an image will be applied to.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>Setting the Frame field to any value other than zero will force the Image
to be drawn only when the drawable's frame matches the specified value.  For
instance, if the drawable container has a Frame setting of 2, and the Image
has a Frame of 1, then the Image graphic will not be drawn as the numbers do
not match.</p>

###SEE ALSO###
Field: Drawable

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frames
Short:  Determines a sequence of frames that an image may be drawn to.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>For long animation sequences you can set the Frames field to determine
individual frames that an image should be drawn to.  For instance, if a
drawable has been assigned 20 animation frames and you want an image to
appear in frames 5, 6, 7, 8, 17 and 19, you would use this frame
specification:</p>

<pre>
   5-8,17,19
</pre>

<p>Individual frame sets are separated by any non-numeric character such as
a comma, and ranges can be specified through the '-' character.</p>

<p>Setting the Frames field will override any setting that you may have
previously set in the Frame field.</p>

###END###

*****************************************************************************/

static ERROR SET_Frames(struct Image *Self, STRING Value)
{
   STRING str;
   LONG frame, end;

   if ((!Value) OR (!(*Value))) return(ERR_Args);

   if (Self->Frames)    { ReleaseMemory(Self->Frames);   Self->Frames = NULL; }
   if (Self->FramesMID) { FreeMemoryID(Self->FramesMID); Self->FramesMID = NULL; }

   /*** Calculate the size of the frame array that we are going to need ***/

   Self->TotalFrames = 0;
   str = Value;
   while (*str) {
      if ((*str >= '0') AND (*str <= '9')) {
         frame = StrToInt(str);
         while ((*str >= '0') AND (*str <= '9')) str++;
         if (frame > Self->TotalFrames) Self->TotalFrames = frame;
      }
      else str++;
   }

   Self->TotalFrames++; /* Add an extra frame because we start from #1, not zero */

   /*** Allocate the frame array and record the sequence ***/

   if (AllocMemory((Self->TotalFrames+1) * sizeof(WORD), MEM_DATA|Self->Head.MemFlags, &Self->Frames, &Self->FramesMID) IS ERR_Okay) {
      str = Value;
      while (*str) {
         if ((*str >= '0') AND (*str <= '9')) {
            if ((frame = StrToInt(str)) > 0) {
               while ((*str >= '0') AND (*str <= '9')) str++;
               while ((*str) AND (*str <= 0x20)) str++;
               if (*str IS '-') {
                  while ((*str) AND ((*str < '0') OR (*str > '9'))) str++;
                  if ((end = StrToInt(str)) > 0) {
                     while (frame != end) {
                        Self->Frames[frame] = 1;
                        if (frame < end) frame++;
                        else frame--;
                     }
                     Self->Frames[frame] = 1;
                  }
                  while ((*str >= '0') AND (*str <= '9')) str++;
               }
               else Self->Frames[frame] = 1;
            }
            else while ((*str >= '0') AND (*str <= '9')) str++;
         }
         else str++;
      }

      Self->Frames[Self->TotalFrames] = -1; /* Sequence termination */
   }
   else return(ERR_AllocMemory);

   Self->Frame = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  The height of an image is specified here.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>If an image graphic is to be limited to a specific drawable area, set this
field to limit the clipping height.  A percentage may be specified if the
FD_PERCENT flag is used when setting the field.</p>

###SEE ALSO###
Field: ImageHeight, ImageWidth, Width

###END###

*****************************************************************************/

static ERROR GET_Height(struct Image *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   FLOAT containerheight, YCoord, YOffset;

   if (!(Value->Type & (FD_DOUBLE|FD_LARGE))) {
      return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   }

   if (Value->Type & FD_PERCENTAGE) {
      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Height, FT_FLOAT, &containerheight);
            Value->Double = Self->Height / containerheight * 100;
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else {
         if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
                  (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
            if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
               GetField(drawable, FID_Height, FD_FLOAT, &containerheight);
               if (Self->Dimensions & DMF_FIXED_Y) YCoord = Self->YCoord;
               else YCoord = Self->YCoord * 100 / containerheight;
               if (Self->Dimensions & DMF_FIXED_YOFFSET) YOffset = Self->YOffset;
               else YOffset = Self->YOffset * 100 / containerheight;
               Value->Double = (containerheight - YOffset - YCoord) / containerheight * 100;
               ReleaseObject(drawable);
            }
            else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
         }
         else return(ERR_FieldNotSet);
      }
   }
   else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
      Value->Double = Self->Height;
   }
   else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         GetField(drawable, FID_Height, FT_FLOAT, &containerheight);
         Value->Double = Self->Height * containerheight / 100;
         ReleaseObject(drawable);
      }
      else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
   }
   else {
      if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Height, FD_FLOAT, &containerheight);
            GetFields(Self, FID_YCoord|TFLOAT,  &YCoord,
                            FID_YOffset|TFLOAT, &YOffset,
                            TAGEND);
            Value->Double = containerheight - YOffset - YCoord;
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else return(ERR_FieldNotSet);
   }

   Value->Large = Value->Double;
   return(ERR_Okay);
}

static ERROR SET_Height(struct Image *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Image] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
      Self->Height     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Image] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
      Self->Height     = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ImageHeight
Short:  Defines an image's height in pixels.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>The height of an image can be read and manipulated through this field.
If you set the height to a specific value then the picture will be cropped
if the height of the picture file exceeds it.  However, if combined with
the ENLARGE or SHRINK flags, the height of the image will be stretched to
fit the given value.</p>

###SEE ALSO###
Fields: Height, ImageWidth, Width

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ImageWidth
Short:  Defines an image's width in pixels.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>The width of an image can be read and manipulated through this field.
If you set the width to a specific value then the picture will be cropped
if the width of the picture file exceeds it.  However, if combined with
the ENLARGE or SHRINK flags, the width of the image will be stretched to
fit the given value.</p>

###SEE ALSO###
Fields: Height, ImageHeight, Width

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:     Location
Synonyms: Src
Short:    Identifies the location of the image graphic file (e.g. pcx, jpeg, gif).
Type:     STRING
Status:   Read/Write

###DESCRIPTION###
<p>The location of the picture file that is to be used for the Image must be
specified in this field.  If the location is not specified then the
initialisation process will fail.  The file format must be recognised by at
least one of the Picture classes loaded into the system.</p>

###END###

*****************************************************************************/

static ERROR GET_Location(struct Image *Self, STRING *Value)
{
   if (Self->Location) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else if (!Self->LocationMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->LocationMID, MEM_READ, (void **)&Self->Location) IS ERR_Okay) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Location(struct Image *Self, STRING Value)
{
   LONG i;

   if (Self->Location)    { ReleaseMemory(Self->Location);   Self->Location = NULL; }
   if (Self->LocationMID) { FreeMemoryID(Self->LocationMID); Self->LocationMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Location, &Self->LocationMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Location[i] = Value[i];
         Self->Location[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Mask
Short:  Identifies the colour to use for masking an image (string format).
Type:   STRING
Status: Write

###DESCRIPTION###
<p>If the Image source uses a masking colour to define transparent areas,
you will need to set this or the MaskRGB field so that the Image object
knows what the masking colour is.  The mask must be specified in
hexadecimal or separated-decimal format - for example a pure red mask
would be defined as "#ff0000" or "255,0,0".</p>

###SEE ALSO###
Field: MaskRGB

###END###

*****************************************************************************/

static ERROR SET_Mask(struct Image *Self, STRING Mask)
{
   if (*Mask IS '#') {
      Mask++;
      Self->MaskRGB.Red = ReadByte(Mask);
      if (*Mask) Mask++; if (*Mask) Mask++;
      Self->MaskRGB.Green = ReadByte(Mask);
      if (*Mask) Mask++; if (*Mask) Mask++;
      Self->MaskRGB.Blue = ReadByte(Mask);
   }
   else {
      Self->MaskRGB.Red = StrToInt(Mask);
      while ((*Mask) AND (*Mask != ',')) Mask++; if (*Mask) Mask++;
      Self->MaskRGB.Green = StrToInt(Mask);
      while ((*Mask) AND (*Mask != ',')) Mask++; if (*Mask) Mask++;
      Self->MaskRGB.Blue = StrToInt(Mask);
   }
   Self->Flags |= IMF_MASK;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MaskRGB
Short:  Identifies the colour to use for masking an image (RGB format).
Type:   *RGB
Status: Read/Write

###DESCRIPTION###
<p>If the Image source uses a masking colour to define transparent areas,
you will need to set this or the Mask field so that the Image object
knows what the masking colour is.  You must pass a valid RGB structure to
set the mask in this field - if you would rather use a colour string, set
the Mask field instead.</p>

###SEE ALSO###
Field: Mask

###END###

*****************************************************************************/

static ERROR GET_MaskRGB(struct Image *Self, struct RGB *RGB)
{
   RGB->Red   = Self->MaskRGB.Red;
   RGB->Green = Self->MaskRGB.Green;
   RGB->Blue  = Self->MaskRGB.Blue;
   return(ERR_Okay);
}

static ERROR SET_MaskRGB(struct Image *Self, struct RGB *RGB)
{
   Self->MaskRGB.Red   = RGB->Red;
   Self->MaskRGB.Green = RGB->Green;
   Self->MaskRGB.Blue  = RGB->Blue;
   Self->Flags |= IMF_MASK;
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Opacity
Short:  Determines the level of translucency applied to an image.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>This field determines the translucency level of an image.  The default
setting is 100%, which means that the image will be solid.  Any other value
that you set here will alter the impact of an image over the destination
drawable.  High values will retain the boldness of the image, while low
values can render it close to invisible.</p>

<p>Please note that the use of translucency will always have an impact on the
time it normally takes to draw an image.  The use of translucency also
requires that the drawable area is buffered, as read access is required to
perform the blending algorithm.</p>

###END###

****************************************************************************/

static ERROR GET_Opacity(struct Image *Self, FLOAT *Value)
{
   *Value = Self->Opacity * 100 / 255;
   return(ERR_Okay);
}

static ERROR SET_Opacity(struct Image *Self, FLOAT *Value)
{
   if (*Value < 0) Self->Opacity = 0;
   else if (*Value > 100) Self->Opacity = 255;
   else Self->Opacity = *Value * 255 / 100;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   PixelSize
Short:  Reflects the overall pixel size of the image at its widest point.
Type:   LONG
Status: Get

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR GET_PixelSize(struct Image *Self, LONG *Value)
{
   if (Self->ImageWidth > Self->ImageHeight) *Value = Self->ImageWidth;
   else *Value = Self->ImageHeight;
   return(ERR_Okay);
}
/*****************************************************************************

###FIELD###
Name:   Tile
Short:  Set this field to TRUE to turn on image tiling (wallpaper).
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>If you want to tile the Image within its container (also known as the
wallpaper effect), set this field to TRUE.</p>

###END###

*****************************************************************************/

static ERROR GET_Tile(struct Image *Self, LONG *Value)
{
   if (Self->Flags & IMF_TILE) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Tile(struct Image *Self, LONG *Value)
{
   if (*Value IS TRUE) Self->Flags |= IMF_TILE;
   else Self->Flags &= ~IMF_TILE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  The width of an image's drawable area is specified here.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>If an image graphic is to be limited to a specific drawable area, set this
field to limit the clipping width.  A percentage may be specified if the
FD_PERCENT flag is used when setting the field.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

static ERROR GET_Width(struct Image *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   FLOAT ContainerWidth, XCoord, XOffset;

   if (!(Value->Type & (FD_DOUBLE|FD_LARGE))) {
      return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   }

   if (Value->Type & FD_PERCENTAGE) {
      if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_FIXED_WIDTH) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width / ContainerWidth * 100;
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else {
         if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
                  (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
            if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
               GetField(drawable, FID_Width, FD_FLOAT, &ContainerWidth);
               if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
               else XCoord = Self->XCoord * 100 / ContainerWidth;
               if (Self->Dimensions & DMF_FIXED_XOFFSET) XOffset = Self->XOffset;
               else XOffset = Self->XOffset * 100 / ContainerWidth;
               Value->Double = (ContainerWidth - XOffset - XCoord) * 100 / ContainerWidth;
               ReleaseObject(drawable);
            }
            else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
         }
         else return(ERR_FieldNotSet);
      }
   }
   else if (Self->Dimensions & DMF_FIXED_WIDTH) {
      Value->Double = Self->Width;
   }
   else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         GetField(drawable, FID_Width, FT_FLOAT, &ContainerWidth);
         Value->Double = Self->Width * ContainerWidth / 100;
         ReleaseObject(drawable);
      }
      else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
   }
   else {
      if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = ContainerWidth * Self->XCoord / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - XCoord - Self->XOffset;
            else Value->Double = ContainerWidth - XCoord - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else return(ERR_Failed);
   }

   Value->Large = Value->Double;
   return(ERR_Okay);
}

static ERROR SET_Width(struct Image *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Image] A width setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
      Self->Width      = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Image] A width setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
      Self->Width      = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  Defines the horizontal position of an image.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of an image can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Fields: Align, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Image *Self, struct Variable *Value)
{
   struct Render *drawable;
   DOUBLE value;

   if (AccessObject(Self->DrawableID, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
      value = GetX(Self, drawable);
      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / drawable->Width;
      ReleaseObject(drawable);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Image *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Defines the horizontal offset of an image.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the image will be drawn
from that X coordinate up to the width of the container, minus the value
given in the XOffset.  This means that the width of the image is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the image will be drawn at an X coordinate calculated from the formula
"XCoord = ContainerWidth - ImageWidth - XOffset".</p>

###SEE ALSO###
Fields: Align, XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Image *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_XOFFSET) value = Self->XOffset;
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         value = Self->XOffset * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_X) value = containerwidth - (Self->XCoord + width);
         else value = containerwidth - ((Self->XCoord * containerwidth / 100) + width);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Image *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XOffset = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XOffset = Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  Defines the vertical position of an image.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of an image can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Fields: Align, XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_Y(struct Image *Self, struct Variable *Value)
{
   struct Render *drawable;
   LONG value;

   if (AccessObject(Self->DrawableID, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
      value = GetY(Self, drawable);
      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / drawable->Height;
      ReleaseObject(drawable);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Image *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Defines the vertical offset of an image.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the image will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the Image is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the image will be drawn at a Y coordinate calculated from the formula
"YCoord = ContainerHeight - ImageHeight - YOffset".</p>

###SEE ALSO###
Fields: Align, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Image *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   value = 0;
   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_YOFFSET) value = Self->YOffset;
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         value = Self->YOffset * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_Y) value = containerheight - (Self->YCoord + height);
         else value = containerheight - ((Self->YCoord * containerheight / 100) + height);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Image *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->YOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************/

static LONG GetX(struct Image *Self, struct Render *Drawable)
{
   LONG width;

   if (Self->Dimensions & DMF_FIXED_X) return(Self->XCoord);
   else if (Self->Dimensions & DMF_RELATIVE_X) {
      return(Self->XCoord * Drawable->Width / 100);
   }
   else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
            (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
      else width = (Drawable->Width * Self->Width) / 100;
         
      if (Self->Dimensions & DMF_FIXED_XOFFSET) return(Drawable->Width - width - Self->XOffset);
      else return(Drawable->Width - width - ((Drawable->Width * Self->XOffset) / 100));
   }
   else return(0);
}

static LONG GetY(struct Image *Self, struct Render *Drawable)
{
   LONG height;

   if (Self->Dimensions & DMF_FIXED_Y) return(Self->YCoord);
   else if (Self->Dimensions & DMF_RELATIVE_Y) {
      return(Self->YCoord * Drawable->Height / 100);
   }
   else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
            (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
      else height = (Drawable->Height * Self->Height) / 100;
         
      if (Self->Dimensions & DMF_FIXED_YOFFSET) return(Drawable->Height - height - Self->YOffset);
      else return(Drawable->Height - height - ((Drawable->Height * Self->YOffset) / 100));
   }
   else return(0);
}

static LONG GetHeight(struct Image *Self, struct Render *Drawable)
{
   LONG y;

   if (Self->Dimensions & DMF_FIXED_HEIGHT) return(Self->Height);
   else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) return((Self->Height * Drawable->Height) / 100);
   else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
            (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
        
     if (Self->Dimensions & DMF_FIXED_Y) y = Self->YCoord;
     else y = (Drawable->Height * Self->YCoord) / 100;
        
     if (Self->Dimensions & DMF_FIXED_YOFFSET) return(Drawable->Height - y - Self->YOffset);
     else return(Drawable->Height - y - (Drawable->Height * Self->YOffset / 100));
   }
   else return(0);
}

static LONG GetWidth(struct Image *Self, struct Render *Drawable)
{
   LONG x;

   if (Self->Dimensions & DMF_FIXED_WIDTH) return(Self->Width);
   else if (Self->Dimensions & DMF_RELATIVE_WIDTH) return((Self->Width * Drawable->Width) / 100);
   else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
            (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
        
     if (Self->Dimensions & DMF_FIXED_X) x = Self->XCoord;
     else x = (Drawable->Width * Self->XCoord) / 100;
        
     if (Self->Dimensions & DMF_FIXED_XOFFSET) return(Drawable->Width - x - Self->XOffset);
     else return(Drawable->Width - x - (Drawable->Width * Self->XOffset / 100));
   }
   else return(0);
}
