#ifndef  MODULES_FILESYSTEM_H
#define  MODULES_FILESYSTEM_H

/*
**   filesystem.h
**
**   (C) Copyright 1996-2004 Rocklyte Systems.
**       All Rights Reserved.
*/

#include <pandora/files/assign.h>
#include <pandora/files/file.h>

#define MODVERSION_FILESYSTEM ((FLOAT)1.1)

#define ASSIGN_REPLACE  0x00000001
#define ASSIGN_PRIORITY 0x00000002
#define ASSIGN_HIDDEN   0x00000004
#define ASSIGN_SAVE     0x00000008
#define ASSIGN_APPEND   0 /* Default */

#define RSF_NOFILECHECK   0x00000001
#define RSF_CHECKRESERVED 0x00000002
#define RSF_APPROXIMATE   0x00000004
#define RSF_WILDCARDS     0x00000008
#define RSF_CHECKVIRTUAL  RSF_CHECKRESERVED

#define DEVICE_COMPACTDISC 0x0000000000000001  /* Compact disc style device */
#define DEVICE_HARDDISK    0x0000000000000002  /* Hard disk style device */
#define DEVICE_FLOPPYDISK  0x0000000000000004  /* Floppy disk style device */
#define DEVICE_READ        0x0000000000000008  /* Device is readable */
#define DEVICE_WRITE       0x0000000000000010  /* Device is writeable */
#define DEVICE_REMOVEABLE  0x0000000000000020  /* Device media is removeable from the hardware */
#define DEVICE_SOFTWARE    0x0000000000000040  /* Device is virtual/software defined */
#define DEVICE_NETWORK     0x0000000000000080  /* Device represents a network link */
#define DEVICE_TAPE        0x0000000000000100  /* Tape/Stream style device */
#define DEVICE_PRINTER     0x0000000000000200
#define DEVICE_SCANNER     0x0000000000000400
#define DEVICE_TEMPORARY   0x0000000000000800  /* All storage is temporary */
#define DEVICE_MEMORY      0x0000000000001000  /* Device is RAM, ROM, WORM, NVRAM or other form of memory */
#define DEVICE_MODEM       0x0000000000002000

#define RDF_SIZE        0x00000001    /* Get file size */
#define RDF_DATE        0x00000002    /* Get file/directory date */
#define RDF_PERMISSIONS 0x00000004    /* Get permission/security information */
#define RDF_FILES       0x00000008    /* Read all files in the directory */
#define RDF_DIRECTORIES 0x00000010    /* Read all folders/assigns in the directory */
#define RDF_ASSIGN      0x00000020    /* Feedback only - indicates an assignment */
#define RDF_SORT        0x00000040    /* Sort the results by name (ascending) */
#define RDF_LINK        0x00000080    /* File/Dir is actually a link to another location */
#define RDF_TAGS        0x00000100    /* Retrieve file-specific information tags */
#define RDF_HIDDEN      0x00000200    /* File/Dir is hidden */
#define RDF_QUALIFY     0x00000400    /* Fully-qualify paths and device names */

#define RDF_READALL     (RDF_SIZE|RDF_DATE|RDF_PERMISSIONS|RDF_FILES|RDF_DIRECTORIES)
#define RDF_TIME        RDF_DATE
#define RDF_FOLDERS     RDF_DIRECTORIES
#define RDF_FILE        RDF_FILES
#define RDF_DIRECTORY   RDF_DIRECTORIES
#define RDF_QUALIFIED   RDF_QUALIFY

enum {
   VAS_DEREGISTER=1,
   VAS_SCANDIR,
   VAS_DELETE,
   VAS_RENAME,
   VAS_READDIR,
   VAS_OPENDIR,
   VAS_CLOSEDIR,
   VAS_TESTLOCATION
};

enum {
   AST_Path=1,
   AST_Name,
   AST_Flags,
   AST_Icon,
   AST_Comment,
   AST_Label,
   AST_Device,
   AST_DevicePath
};

#define AST_PATH       AST_Path
#define AST_NAME       AST_Name
#define AST_FLAGS      AST_Flags
#define AST_ICON       AST_Icon
#define AST_COMMENT    AST_Comment
#define AST_LABEL      AST_Label
#define AST_DEVICE     AST_Device
#define AST_DEVICEPATH AST_DevicePath

struct FileInfo {
   struct FileInfo *Next;
   STRING Name;
   STRING *Tags;
   LONG   Flags;
   LONG   Permissions;
   LARGE  Size;
   struct FileTime Time;
   LARGE  TimeStamp;
   LONG   UserID;
   LONG   GroupID;
};

struct DirInfo {
   /*** Important that this structure identical to FileInfo, except for private variables at the end ***/
   struct FileInfo *Next;
   STRING Name;
   STRING *Tags;
   LONG   Flags;
   LONG   Permissions;
   LARGE  Size;
   struct FileTime Time;
   LARGE  TimeStamp;
   LONG   UserID;
   LONG   GroupID;
   struct {
      APTR   Handle;        /* Directory handle.  If virtual, may store a private data address */
      STRING Path;          /* Original directory location string */
      STRING ResolvedPath;  /* Resolved directory location */
      LONG   Flags;         /* OpenDirectory() flags */
      WORD   Index;         /* Current index within the directory */
      WORD   ResolveLen;    /* Byte length of ResolvedPath */
      WORD   VirtualIndex;  /* Set to an index that refers to a virtual device */
      WORD   Total;         /* Total number of items in the directory */
   } Dir;
};

struct DeviceInfo {
   LARGE Flags;            /* Flags identifying the type of media */
   DOUBLE DeviceSize;      /* Size of the device */
   DOUBLE BytesAvailable;  /* Bytes available to the user */
   DOUBLE BytesUsed;       /* Bytes already used */
};

struct FileFeedback {
   LONG   FeedbackID;         /* Set to one of the FDB integers */
   STRING Location;           /* Location of the file being processed */
   union {
      STRING Destination;        /* Destination file/path if moving or copying */
      STRING Dest;
   };
   LARGE  Size;               /* Size of the file */
   LARGE  Position;           /* Current seek position within the file if moving or copying */
   APTR   User;               /* User data, set by FileFeedback() */
   UBYTE  Reserved[32];       /* Reserved in case of future expansion */
};

#define FDB_MoveFile   1
#define FDB_CopyFile   2
#define FDB_DeleteFile 3

/*** Return codes available to the feedback routine ***/

#define FFR_Okay   1
#define FFR_Skip   2
#define FFR_Abort  3

#ifndef LOCAL_FILESYSTEM

struct FileSystemBase {
   ERROR (*AnalyseLocation)(STRING, LONG *);
   ERROR (*MakeDirectory)(STRING, LONG);
   ERROR (*MoveLocation)(STRING, STRING);
   ERROR (*ResolveAssign)(STRING, STRING *, LONG);
   ERROR (*SetAssignTags)(LONG, ...);
   ERROR (*DeleteAssign)(STRING);
   ERROR (*VirtualAssignTags)(STRING, ...);
   ERROR (*CopyLocation)(STRING, STRING);
   ERROR (*GetDeviceInfo)(STRING, struct DeviceInfo *, LONG);
   ERROR (*ReadDirectory)(STRING, LONG, struct FileInfo **);
   STRING (*GetFileIcon)(STRING);
   ERROR (*DeleteFile)(STRING);
   ERROR (*GetFileInfo)(STRING, struct FileInfo *, LONG);
   ERROR (*SaveObjectToFile)(OBJECTPTR, STRING, LONG);
   void  (*CloseDirectory)(struct DirInfo *);
   ERROR (*OpenDirectory)(STRING, LONG, struct DirInfo **);
   ERROR (*ScanDirectory)(struct DirInfo *);
   ERROR (*IdentifyFile)(STRING Location, CLASSID *ClassID, CLASSID *SubClassID);
   LONG  (*FileCaseSensitive)(LONG);
   ERROR (*CreateLink)(STRING, STRING); /* V1.1 (Kernel 3.5) */
   ERROR (*FileFeedback)(APTR, APTR, LONG); /* V1.1 (Kernel 3.5) */
   void  (*ForcePermissions)(LONG, LONG, LONG); /* V1.2 (Kernel 4.1) */
   STRING (*ConvertUserID)(LONG); /* V1.2 */
   STRING (*ConvertGroupID)(LONG); /* V1.2 */
};

#define SetAssign               (FileSystemBase->SetAssignTags)
#define AnalyseLocation(a,b)    (FileSystemBase->AnalyseLocation(a,b))
#define AssignList(a)           (FileSystemBase->AssignList(a))
#define CloseDirectory(a)       (FileSystemBase->CloseDirectory(a))
#define ConvertUserID(a)        (FileSystemBase->ConvertUserID(a))
#define ConvertGroupID(a)       (FileSystemBase->ConvertGroupID(a))
#define CopyLocation(a,b)       (FileSystemBase->CopyLocation(a,b))
#define CreateLink(a,b)         (FileSystemBase->CreateLink(a,b))
#define DeleteAssign(a)         (FileSystemBase->DeleteAssign(a))
#define DeleteFile(a)           (FileSystemBase->DeleteFile(a))
#define FileCaseSensitive(a)    (FileSystemBase->FileCaseSensitive(a))
#define FileFeedback(a,b,c)     (FileSystemBase->FileFeedback(a,b,c))
#define ForcePermissions(a,b,c) (FileSystemBase->ForcePermissions(a,b,c))
#define GetDeviceInfo(a,b,c)    (FileSystemBase->GetDeviceInfo(a,b,c))
#define GetFileIcon(a)          (FileSystemBase->GetFileIcon(a))
#define GetFileInfo(a,b,c)      (FileSystemBase->GetFileInfo(a,b,c))
#define IdentifyFile(a,b,c)     (FileSystemBase->IdentifyFile(a,b,c))
#define MakeDirectory(a,b)      (FileSystemBase->MakeDirectory(a,b))
#define MoveLocation(a,b)       (FileSystemBase->MoveLocation(a,b))
#define OpenDirectory(a,b,c)    (FileSystemBase->OpenDirectory(a,b,c))
#define ReadDirectory(a,b,c)    (FileSystemBase->ReadDirectory(a,b,c))
#define ResolveAssign(a,b,c)    (FileSystemBase->ResolveAssign(a,b,c))
#define SaveObjectToFile(a,b,c) (FileSystemBase->SaveObjectToFile(a,b,c))
#define ScanDirectory(a)        (FileSystemBase->ScanDirectory(a))
#define VirtualAssign           (FileSystemBase->VirtualAssignTags)

#endif /* LOCAL_FILESYSTEM */

#endif /* MODULES_FILESYSTEM_H */
